from django.contrib import admin
from django.db.utils import IntegrityError
from django.http import HttpRequest

from .models import (
    CspReport,
    CspReportBlacklist,
    CspReportQuerySet,
    CspRule,
    CspRuleQuerySet,
    convert_report,
)
from .utils import strip_path


@admin.register(CspRule)
class CspRuleAdmin(admin.ModelAdmin):
    list_display = ("directive", "value", "_enabled", "modified_at")
    list_filter = ("created_at", "modified_at", "directive", "enabled")
    search_fields = ("value",)
    ordering = ("-modified_at",)
    actions = [
        "enable_selected_rules",
        "disable_selected_rules",
        "strip_selected_rules",
    ]

    @admin.display(boolean=True)
    def _enabled(self, obj: CspRule) -> bool:
        return obj.enabled

    def clear_cache(self) -> None:
        from .policy import clear_cache as clear_csp_cache

        clear_csp_cache()

    @admin.action(description="Enable selected CSP rules")
    def enable_selected_rules(
        self, request: HttpRequest, queryset: CspRuleQuerySet
    ) -> None:
        count = queryset.update(enabled=True)
        self.clear_cache()
        self.message_user(request, f"Enabled {count} rules.")

    @admin.action(description="Disable selected CSP rules")
    def disable_selected_rules(
        self, request: HttpRequest, queryset: CspRuleQuerySet
    ) -> None:
        count = queryset.update(enabled=False)
        self.clear_cache()
        self.message_user(request, f"Disabled {count} rules.")

    @admin.action(description="Strip path from selected CSP rules")
    def strip_selected_rules(
        self, request: HttpRequest, queryset: CspReportQuerySet
    ) -> None:
        """Strip paths off selected rules."""
        stripped = 0
        ignored = 0
        deleted = 0
        for rule in queryset:
            stripped_value = strip_path(rule.value)
            if stripped_value == rule.value:
                ignored += 1
            else:
                rule.value = stripped_value
                try:
                    rule.save(update_fields=["value"])
                except IntegrityError:
                    # we have a duplicate rule - because we are
                    # stripping off the path multiple rules may now
                    # clash (same origin), and so we delete the
                    # duplicates.
                    rule.delete()
                    deleted += 1
                else:
                    stripped += 1
        if stripped:
            self.message_user(
                request, f"Successfully stripped {stripped} rules.", "success"
            )
        if ignored:
            self.message_user(request, f"Ignored {ignored} unchanged rules.", "success")
        if deleted:
            self.message_user(request, f"Deleted {deleted} duplicate rules.", "error")


@admin.register(CspReport)
class CspReportAdmin(admin.ModelAdmin):
    list_display = (
        "effective_directive",
        "blocked_uri",
        "request_count",
        "last_updated_at",
    )
    readonly_fields = (
        "document_uri",
        "effective_directive",
        "disposition",
        "blocked_uri",
        "created_at",
        "last_updated_at",
        "request_count",
    )
    list_filter = ("effective_directive", "last_updated_at")
    actions = ("add_rule", "add_to_blacklist")

    @admin.action(description="Add new CSP rule for selected violations.")
    def add_rule(self, request: HttpRequest, queryset: CspReportQuerySet) -> None:
        created: list[CspRule] = []
        duplicates = 0
        for report in queryset:
            if rule := convert_report(report, enable=True):
                created.append(rule)
            else:
                duplicates += 1
        if created:
            self.message_user(request, f"Created {len(created)} new rules.", "success")
        if duplicates:
            self.message_user(request, f"Ignored {duplicates} duplicates.", "warning")

    @admin.action(description="Blacklist selected violations.")
    def add_to_blacklist(
        self, request: HttpRequest, queryset: CspReportQuerySet
    ) -> None:
        blacklisted = 0
        duplicates = 0
        for report in queryset:
            obj, created = CspReportBlacklist.objects.get_or_create(
                directive=report.effective_directive, blocked_uri=report.blocked_uri
            )
            report.delete()
            if created:
                blacklisted += 1
            else:
                duplicates += 1
        if blacklisted:
            self.message_user(request, f"Blacklisted {blacklisted} reports.", "success")
        if duplicates:
            self.message_user(request, f"Ignored {duplicates} duplicates.", "warning")


@admin.register(CspReportBlacklist)
class CspReportBlacklistAdmin(admin.ModelAdmin):
    list_display = (
        "directive",
        "blocked_uri",
    )
    list_filter = ("directive",)



# Generated by Django 4.1.7 on 2023-02-16 18:24

import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("csp", "0002_cspreportblacklist"),
    ]

    operations = [
        migrations.AddField(
            model_name="csprule",
            name="created_at",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AddField(
            model_name="csprule",
            name="modified_at",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
    ]



# Generated by Django 4.1.3 on 2022-12-10 12:38

from django.db import migrations, models

from csp.models import DirectiveChoices


class Migration(migrations.Migration):
    dependencies = [
        ("csp", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="CspReportBlacklist",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "directive",
                    models.CharField(
                        choices=DirectiveChoices.choices,
                        max_length=50,
                    ),
                ),
                ("blocked_uri", models.URLField()),
            ],
            options={
                "verbose_name": "CSP Blacklist",
                "verbose_name_plural": "CSP Blacklist",
                "ordering": ["directive", "blocked_uri"],
                "unique_together": {("directive", "blocked_uri")},
            },
        ),
    ]



from django.apps import AppConfig


class CSPTrackerConfig(AppConfig):
    name = "csp"
    verbose_name = "CSP Tracker"
    default_auto_field = "django.db.models.BigAutoField"

    def ready(self) -> None:
        from . import signals  # noqa

        self.reset()
        super().ready()

    def reset(self) -> None:
        """Ensure that cache is cleared on startup."""
        from .policy import clear_cache

        clear_cache()



from __future__ import annotations

from copy import deepcopy
from typing import Callable, TypeAlias

from django.conf import settings
from django.http import HttpRequest, HttpResponse

PolicyType: TypeAlias = dict[str, list[str]]


# If False then the middleware is disabled completely
CSP_ENABLED = bool(getattr(settings, "CSP_ENABLED", False))


# If True then set the report-only attr on the CSP
CSP_REPORT_ONLY = bool(getattr(settings, "CSP_REPORT_ONLY", True))

# === reporting ===
#
# This is complicated - the reporting process is in a state of flux, and
# there are three scenarios to support in order to handle today's
# browsers and the newer, propoosed standards.
#
# Phase 1 (today): reporting uses the "report-uri <uri>"  CSP directive
#
# Phase 2 (today (partial)): reporting uses the "report-to <endpoint>"
# SCP directive, which in turn relies on the browser reporting API.
#
# Phase 2.1 the reporting API uses the "Report-To" HTTP header
#
# Phase 2.2 the reporting API uses the new "Reporting-Endpoints" header.
#
# We make no attempt to control this - but we do provide support for the
# reporting API headers - if you supply them, we'll add them.
#
# See https://developer.chrome.com/blog/reporting-api-migration/#migration-steps-for-csp-reporting  # noqa: E501
# for a migration plan, and https://www.w3.org/TR/reporting-1/ for the spec.


# The Report-To header value - if supplied it will be added to the response -
# and you can then add a "report-to: <endpoint>" directive to the CSP.
# https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Security-Policy/report-to  # noqa: E501
REPORT_TO_HEADER = getattr(settings, "REPORT_TO_HEADER", None)


# The Reporting-Endpoints value - if supplied it will be added to the response -
# and you can then add a "report-to: <endpoint>" directive to the CSP.
# https://developer.chrome.com/blog/reporting-api-migration/#migration-steps-for-csp-reporting  # noqa: E501
REPORTING_ENDPOINTS_HEADER = getattr(settings, "REPORTING_ENDPOINTS_HEADER", None)


# Value 0..1 - used to tune the percentage of responses that get the
# report-uri valuable if the reporting is too noisy. Set to 0.0 to
# disable report-uri completely, or 1.0 to include it on all responses.
CSP_REPORT_SAMPLING = float(getattr(settings, "CSP_REPORT_SAMPLING", 1.0))


# Value 0..1 - used to throttle report-uri requests. The report-uri is
# an open endpoint that accepts JSON payloads - and as such represents a
# DOS vulnerability. Use this to throw away a percentage of reports
# received without attempting to process them. Set to 1.0 to ignore all
# inbound reports.
CSP_REPORT_THROTTLING = float(getattr(settings, "CSP_REPORT_THROTTLING", 0.0))


# dict to downgrade unsupported directives when converting to rules,
# e.g. if the violation from Chrome is "script-src-elem", which is not
# universally supported, then convert it to "script-src" on the fly.
CSP_REPORT_DIRECTIVE_DOWNGRADE: dict[str, str] = getattr(
    settings,
    "CSP_REPORT_DIRECTIVE_DOWNGRADE",
    {
        "script-src-elem": "script-src",
        "script-src-attr": "script-src",
        "style-src-elem": "style-src",
        "style-src-attr": "style-src",
    },
)


# Name of the header value to use based on CSP_REPORT_ONLY
CSP_RESPONSE_HEADER = {
    True: "Content-Security-Policy-Report-Only",
    False: "Content-Security-Policy",
}[CSP_REPORT_ONLY]


# cache timeout in seconds - defaults to one hour
CSP_CACHE_TIMEOUT = int(getattr(settings, "CSP_CACHE_TIMEOUT", 3600))


# default process_request func
def _process_request(request: HttpRequest) -> bool:
    return True


# default process_response funct
def _process_response(response: HttpResponse) -> bool:
    return "text/html" in response.headers.get("content-type", "")


# True if the request should have the header; defaults to HTML pages only.
process_request: Callable[[HttpRequest], bool] = getattr(
    settings, "CSP_FILTER_REQUEST_FUNC", _process_request
)


process_response: Callable[[HttpResponse], bool] = getattr(
    settings, "CSP_FILTER_RESPONSE_FUNC", _process_response
)


# Default rules from https://content-security-policy.com/
def get_default_rules() -> PolicyType:
    if defaults := getattr(settings, "CSP_DEFAULTS", None):
        # if we don't return a deepcopy alterations to the
        # dictionary will update the lists, meaning that
        # values get stuck.
        return deepcopy(defaults)
    return {
        "default-src": ["'none'"],
        "base-uri": ["'self'"],
        "connect-src": ["'self'"],
        "form-action": ["'self'"],
        "font-src": ["'self'"],
        "img-src": ["'self'"],
        "script-src": ["'self'"],
        "style-src": ["'self'"],
        "report-uri": ["{report_uri}"],
    }


def get_default_rules_expanded() -> list[tuple[str, str]]:
    """
    Return rules as a list of (directive, rule) tuples.

    This format is less intuitive than the dict, but it's compatible
    with the ".values('directive', 'value')" output from the database,
    which makes it easier to blend in.

    """
    # Ask ChatGPT: how can I expand a dictionary of lists in python to a
    # list of tuples?
    return [(k, v) for k, l in get_default_rules().items() for v in l]



from django.urls import path

from .views import csp_diagnostics, report_uri

app_name = "csp"

urlpatterns = [
    path("report-uri/", report_uri, name="report_uri"),
    path("diagnostics/", csp_diagnostics, name="csp_diagnostics"),
]



import json
import logging
import random
from typing import Callable, TypeAlias

from django.contrib.auth.decorators import user_passes_test
from django.db.utils import IntegrityError
from django.http import HttpRequest, HttpResponse, HttpResponseBadRequest
from django.shortcuts import render
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_http_methods
from pydantic import ValidationError

from .blacklist import is_blacklisted
from .models import CspReport, CspRule, ReportData
from .policy import get_csp
from .settings import (
    CSP_REPORT_DIRECTIVE_DOWNGRADE,
    CSP_REPORT_THROTTLING,
    get_default_rules,
)

logger = logging.getLogger(__name__)

SimpleViewType: TypeAlias = Callable[[HttpRequest], HttpResponse]


def throttle_view(func: SimpleViewType) -> SimpleViewType:
    def wrapper(request: HttpRequest) -> HttpResponse:
        # CSP_REPORT_THROTTLING is a float 0..1 - if we're below the value,
        # then respond immediately without attempting to process the
        # payload.
        if random.random() < CSP_REPORT_THROTTLING:  # noqa: S311
            return HttpResponse()
        return func(request)

    return wrapper


def _validation_error(ex: ValidationError) -> str:
    """
    Parse Pydantic error into output message.

    This message goes back to the requestor, so we don't want the
    full Pydantic error - just the list of invalid fields.

    See https://docs.pydantic.dev/usage/models/#error-handling

    """
    try:
        return ", ".join([str(e["loc"][0]) for e in ex.errors()])
    except (IndexError, KeyError):
        logger.exception("Unparseable Pydantic error: %s", ex.json())
        return "unknown error"


@csrf_exempt
@require_http_methods(["POST"])
@throttle_view
def report_uri(request: HttpRequest) -> HttpResponse:
    # {
    #     'csp-report': {
    #         'document-uri': 'http://127.0.0.1:8000/test/',
    #         'referrer': '',
    #         'violated-directive': 'img-src',
    #         'effective-directive': 'img-src',
    #         'original-policy': "default-src https:; img-src 'self';",
    #         'disposition': 'enforce',
    #         'blocked-uri': 'https://example.com',
    #         'line-number': 8,
    #         'source-file': 'http://127.0.0.1:8000/test/',
    #         'status-code': 200,
    #         'script-sample': ''
    #     }
    # }
    request_body = request.body.decode()
    user_agent = request.headers.get("User-Agent", "missing User-Agent")

    def _bad_request(msg: str) -> HttpResponseBadRequest:
        logger.debug(msg)
        logger.debug("CSP reporting User-Agent: %s", user_agent)
        logger.debug("CSP report payload:\n%s", request_body)
        return HttpResponseBadRequest(msg)

    try:
        data = json.loads(request_body)
        report = ReportData(**data["csp-report"])
        if is_blacklisted(report):
            logger.debug("Ignoring blacklisted CSP report")
            return HttpResponse()
        CspReport.objects.save_report(report)
    except json.decoder.JSONDecodeError:
        return _bad_request("Invalid CSP report - must contain valid JSON.")
    except KeyError:
        return _bad_request("Invalid CSP report - must contain 'csp-report'")
    except ValidationError as ex:
        return _bad_request(
            f"Invalid CSP report - report data is invalid: {_validation_error(ex)}"
        )
    except (IntegrityError, CspReport.DoesNotExist, CspReport.MultipleObjectsReturned):
        logger.exception("Error saving CspReport")
        return HttpResponse()
    return HttpResponse(status=201, content_type="application/json")


@user_passes_test(lambda user: user.is_staff)
@require_http_methods(["GET"])
def csp_diagnostics(request: HttpRequest) -> HttpResponse:
    default_rules = get_default_rules()
    extra_rules = list(CspRule.objects.enabled().directive_values())
    csp_list = [x.strip() for x in get_csp(request, True).split(";")]
    return render(
        request,
        "csp/diagnostics.txt",
        {
            "default_rules": default_rules,
            "extra_rules": extra_rules,
            "downgrades": CSP_REPORT_DIRECTIVE_DOWNGRADE,
            "csp": csp_list,
        },
        content_type="text/plain",
    )



# <scheme>://<netloc>/<path>;<params>?<query>#<fragment>
from urllib.parse import urlparse, urlunparse


def strip_fragment(url: str) -> str:
    """Strip the fragment from a url."""
    scheme, netloc, path, params, query, _ = urlparse(url)
    if not scheme:
        return url
    return urlunparse((scheme, netloc, path, params, query, ""))


def strip_query(url: str) -> str:
    """Strip the query, fragment from a url."""
    return url.split("?")[0] if url else url


def strip_path(url: str) -> str:
    """Strip the path, query, fragment from a url."""
    scheme, netloc, _, _, _, _ = urlparse(url)
    if not scheme:
        return url
    # scheme only - must end in ":", which urlunparse will munge
    if scheme and not netloc:
        return url
    return urlunparse((scheme, netloc, "", "", "", ""))



from __future__ import annotations

import logging

from django.core.cache import cache

from .models import CspReportBlacklist, ReportData
from .settings import CSP_CACHE_TIMEOUT, PolicyType

logger = logging.getLogger(__name__)

CACHE_KEY_BLACKLIST = "csp::blacklist"


def clear_cache() -> None:
    """Clear the cached blacklist."""
    logger.debug("Clearing CSP blacklist cache")
    cache.delete(CACHE_KEY_BLACKLIST)


def refresh_cache() -> None:
    """Refresh the cached blacklist."""
    logger.debug("Refreshing CSP blacklist cache")
    blacklist = CspReportBlacklist.objects.all().as_dict()
    cache.set(CACHE_KEY_BLACKLIST, blacklist, CSP_CACHE_TIMEOUT)


def get_blacklist(directive: str) -> PolicyType:
    """Fetch the CSP blacklist from cache."""
    # must check against None as default blacklist is falsey {}
    if (blacklist := cache.get(CACHE_KEY_BLACKLIST)) is not None:
        return blacklist.get(directive, [])
    refresh_cache()
    return get_blacklist(directive)


def is_blacklisted(report: ReportData) -> bool:
    """Return True if the report should be ignored."""
    # blacklist anything that doesn't have an effective_directive
    if not report.effective_directive:
        return True
    blacklisted_sources = get_blacklist(report.effective_directive)
    return any([b for b in blacklisted_sources if report.blocked_uri.startswith(b)])



# Generated by Django 4.1.3 on 2022-12-03 15:38

import django.utils.timezone
from django.db import migrations, models

from csp.models import DirectiveChoices


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="CspRule",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "directive",
                    models.CharField(
                        choices=DirectiveChoices.choices,
                        max_length=50,
                    ),
                ),
                ("value", models.CharField(max_length=255)),
                ("enabled", models.BooleanField(default=False)),
            ],
            options={
                "verbose_name": "CSP Rule",
                "ordering": ["directive", "value"],
                "unique_together": {("value", "directive")},
            },
        ),
        migrations.CreateModel(
            name="CspReport",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("document_uri", models.URLField()),
                ("effective_directive", models.TextField()),
                ("disposition", models.CharField(max_length=12)),
                ("blocked_uri", models.URLField()),
                ("request_count", models.IntegerField(default=0)),
                ("created_at", models.DateTimeField(default=django.utils.timezone.now)),
                (
                    "last_updated_at",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
            ],
            options={
                "verbose_name": "CSP Violation",
                "ordering": ["effective_directive", "blocked_uri"],
                "unique_together": {("effective_directive", "blocked_uri")},
            },
        ),
    ]



from __future__ import annotations

import logging
from collections import defaultdict

from django.core.cache import cache
from django.http import HttpRequest
from django.urls import reverse

from .models import CspRule, DirectiveChoices
from .settings import (
    CSP_CACHE_TIMEOUT,
    CSP_REPORT_DIRECTIVE_DOWNGRADE,
    PolicyType,
    get_default_rules_expanded,
)

logger = logging.getLogger(__name__)

CACHE_KEY_RULES = "csp::rules"


def clear_cache() -> None:
    """Clear the cached CSP."""
    logger.debug("Clearing CSP cache")
    cache.delete(CACHE_KEY_RULES)


def refresh_rules_cache() -> None:
    """Refresh the cached CSP."""
    logger.debug("Refreshing CSP cache")
    policy = build_policy()
    part_one = format_as_csp({k: v for k, v in policy.items() if k != "report-uri"})
    part_two = format_as_csp({k: v for k, v in policy.items() if k == "report-uri"})
    cache.set(CACHE_KEY_RULES, (part_one, part_two), CSP_CACHE_TIMEOUT)


def _dedupe(values: list[str]) -> list[str]:
    return list({CspRule.clean_value(v) for v in values})


def _downgrade(directive: str) -> str:
    """
    Downgrade a directive if configured.

    Some directives have poor support, so we downgrade them on the fly -
    a `script-src-elem` for instance will become `script-src` if defined
    in CSP_REPORT_DIRECTIVE_DOWNGRADE.

    """
    if directive in CSP_REPORT_DIRECTIVE_DOWNGRADE:
        logger.debug("Downgrading directive '%s'", directive)
        return CSP_REPORT_DIRECTIVE_DOWNGRADE[directive]
    return directive


def build_policy() -> PolicyType:
    """
    Build the CSP by combining default settings and CspRules.

    The CSP is a dict of directives and values. This function takes the
    default rules (from settings) and combines them with the CspRules
    from the database to create a cacheable object that represents the
    current state of the CSP.

    As part of building the CSP this function will also dedupe the
    values, and format any special values ('self', 'unsafe-inline',
    etc.) which must be formatted with the single-quotes.

    NB the CSP as cached is not quite complete - if the settings require
    a nonce to be added to any directives then this cannot be cached,
    and so the nonce is applied per-request.

    """
    logger.debug("Building new CSP")

    # return dict of {directive: [values]}
    policy: PolicyType = defaultdict(list)

    def add_directive(directive: str, value: str) -> None:
        directive = _downgrade(directive)
        if directive in DirectiveChoices.values:
            logger.debug('Adding "%s" to directive "%s"', value, directive)
            policy[directive].append(value)
        else:
            logger.debug('Ignoring unknown directive "%s"', directive)

    for directive, value in get_default_rules_expanded():
        add_directive(directive, value)

    # returns list of additional (directive, value) tuples.
    new_rules = CspRule.objects.enabled().directive_values()
    for directive, value in new_rules:
        add_directive(directive, value)

    return {k: _dedupe(v) for k, v in policy.items()}


def format_as_csp(policy: PolicyType) -> str:
    """Convert policty dict into response header string."""
    directives = []
    for directive, values in policy.items():
        if not values:
            continue
        # combine values into a single space-delimited string
        value_str = " ".join(values)
        directives.append(f"{directive} {value_str}")
    # combine directives into a ";" delimited str - the CSP.
    return "; ".join(directives).strip()


def _context(request: HttpRequest) -> dict[str, str]:
    """Return report_uri and nonce."""
    context = {"report_uri": reverse("csp:report_uri")}
    if nonce := getattr(request, "csp_nonce", ""):
        context["nonce"] = f"'nonce-{nonce}'"
    return context


def get_csp(request: HttpRequest, add_report_uri: bool) -> str:
    """Fetch the CSP from the cache, or rebuild if it's missing."""
    if cached_csp := cache.get(CACHE_KEY_RULES):
        logger.debug("Found cached CSP (add report-uri: %s)", add_report_uri)
        csp = "; ".join(cached_csp) if add_report_uri else cached_csp[0]
        return csp.format(**_context(request))
    logger.debug("No cached CSP - rebuilding policy")
    refresh_rules_cache()
    return get_csp(request, add_report_uri)



from django.core.management.base import BaseCommand

from csp.models import CspReport


class Command(BaseCommand):
    help = "Clears out CSP violation reports"

    def handle(self, *args: object, **options: object) -> None:
        count, _ = CspReport.objects.all().delete()
        self.stdout.write(f"Deleted {count} CspReport objects.")



from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver

from .blacklist import clear_cache as clear_blacklist_cache
from .models import CspReportBlacklist, CspRule
from .policy import clear_cache as clear_policy_cache


@receiver([post_save, post_delete], sender=CspRule, dispatch_uid="clear_policy_cache")
def clear_cache_1(sender: type[CspRule], **kwargs: object) -> None:
    clear_policy_cache()


@receiver(
    [post_save, post_delete],
    sender=CspReportBlacklist,
    dispatch_uid="clear_blacklist_cache",
)
def clear_clear_cache_2(sender: type[CspReportBlacklist], **kwargs: object) -> None:
    clear_blacklist_cache()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['csp', 'csp.management', 'csp.management.commands', 'csp.migrations']

package_data = \
{'': ['*'], 'csp': ['fixtures/*', 'templates/csp/*']}

install_requires = \
['django>=3.2,<5.0', 'pydantic>=1,<2']

setup_kwargs = {
    'name': 'django-csp-plus',
    'version': '0.10.0',
    'description': 'CSP tracking and violation report endpoint.',
    'long_description': '# Django CSP Plus\n\nDjango app for building CSP and tracking violations.\n\nThis project is based on the excellent `django-csp` project from MDN,\nwith a couple of alterations:\n\n1. It includes a violation report tracker\n2. It stores rules in a model, so they can be edited at runtime\n\nThe `nonce` pattern has been lifted directly.\n\n## History\n\nThe original reason for forking this from the original was the desire to\nhave the violation reporting with the same Django project as the source\npages. I\'m sure there is / was an excellent reason for not doing so in\nthe original, but it\'s not explained, and Django seems like a great fit\nfor an HTTP endpoint that can parse JSON requests and store the data\nsomewhere.\n\nThe second reason was the experience we had with Sqreen - a fantastic\nsecurity app that we used from their beta launch through to their\nacquisition by Datadog. They have/had a great violation report tool that\nallowed you to see how many violations had occurred, and to\nautomatically add the valid domains to the working CSP, making CSP\ntrivial to manage (and requiring no restarts).\n\nIt felt like this is something we could add to the Django admin\nrelatively easily ("convert this violation report into a rule").\n\nThe final push was the desire to manage the rules at runtime - running a\nlarge commercial site you never quite know what the marketing team has\njust added to the site, and having to redeploy to unblock their new tool\nwas a pain.\n\nWe ended with these requirements:\n\n1. Design time base rules\n2. Runtime configurable rules\n3. Builtin violation reporting\n4. Support for nonces\n5. Ability to exclude specific requests / responses\n\n## Implementation\n\nWe have split the middleware in two - `CspNonceMiddleware`, which adds\nthe `request.csp_nonce` attribute, and `CspHeaderMiddleware`, which adds\nthe header. Most sites will want both, but you can run one without the\nother.\n\nThe baseline, static, configuration of rules is a dict in `settings.py`.\nThis can then be enriched with dynamic rules stored in the `CspRule`\nmodel.\n\nYou can add two special placeholders in the rules: `{nonce}` and\n`{report-uri}`; if present these will be replaced with the current\n`request.csp_nonce` and the local violation report URL on each request.\nThe CSP is cached for all requests with the placeholder text in (so it\'s\nthe same for all users / requests).\n\n### Directives\n\nSome directives are deprecated, and others not-yet implemented. The\ncanonical example is the `style-src-elem` directive (and its `style-`\nand `-attr`) siblings which are _not_ supported by Safari. In order to\nhighlight these the corresponding directive choice labels have been\namended. Treat with caution as setting these attributes may have\nunintended consequences.\n\n#### Downgrading directives\n\nIn some instances you may want to "downgrade" a directive - for instance\nconverting all `script-src-elem` directives to `script-src` (for\ncompatibility reasons). This can be done using the\n`CSP_REPORT_DIRECTIVE_DOWNGRADE` setting.\n\n## Settings\n\n### `CSP_ENABLED`\n\n`bool`, default = `False`\n\nKill switch for the middleware. Defaults to `False` (disabled).\n\n### `CSP_REPORT_DIRECTIVE_DOWNGRADE`\n\n`dict[str, str]`, default =\n```python\n{\n    "script-src-elem": "script-src",\n    "script-src-attr": "script-src",\n    "style-src-elem": "style-src",\n    "style-src-attr": "style-src",\n}\n```\n\nThis is used to transparently "downgrade" any directives to a different\ndirective, and is primarily used for managing compatibility.\n\n### `CSP_REPORT_ONLY`\n\n`bool`, default = `True`\n\nSet to `True` to run in report-only mode. Defaults to `True`.\n\n### `CSP_REPORT_SAMPLING`\n\n`float`, default = `1.0`\n\nFloat (0.0-1.0) - used as a percentage of responses on which to include\nthe `report-uri` directive. This can be used to turn down the noise -\nonce you have a stable CSP there is no point having every single request\ninclude the reporting directive - you need a trickle not a flood.\n\n### `CSP_REPORT_THROTTLING`\n\n`float`, default = `0.0`\n\nFloat (0.0-1.0) - used as a percentage of reporting violation requests\nto throttle (throw away). This is used to control potentially malicious\nviolation reporting. The reporting endpoint is public, and accepts JSON\npayloads, so is open to abuse (sending very large, or malformed JSON)\nand is a potential DOS vulnerability. If you set this value to 1.0 then\nall inbound reporting requests are thrown away without processing. Use\nin extremis.\n\n### `CSP_CACHE_TIMEOUT`\n\n`int`, default = `600`\n\nThe cache timeout for the templated CSP. Defaults to 5 min (600s).\n\n### `CSP_FILTER_REQUEST_FUNC`\n\n`Callable[[HttpRequest], bool]` - defaults to returning `True` for all\nrequests\n\nA callable that takes `HttpRequest` and returns a bool - if False, the\nmiddleware will not add the response header. Defaults to return `True`\nfor all requests.\n\n### `CSP_FILTER_RESPONSE_FUNC`\n\n`Callable[[HttpResponse], bool]` - defaults to `True` for all\n`text/html` responses.\n\nCallable that takes `HttpResponse` and returns a bool - if `False` the\nmiddleware will not add the response header. Defaults to a function that\nfilters only responses with `Content-Type: text/html` - which results in\nstatic content / JSON responses _not_ getting the CSP header.\n\n### `CSP_DEFAULTS`\n\n`dict[str, list[str]]`\n\nThe default (baseline) CSP as a dict of `{directive: values}`. This is\nextended by the runtime rules (i.e. not overwritten). Defaults to:\n\n```python\n{\n    "default-src": ["\'none\'"],\n    "base-uri": ["\'self\'"],\n    "connect-src": ["\'self\'"],\n    "form-action": ["\'self\'"],\n    "font-src": ["\'self\'"],\n    "img-src": ["\'self\'"],\n    "script-src": ["\'self\'"],\n    "style-src": ["\'self\'"],\n    "report-uri": ["{report_uri}"],\n}\n```\n\nNote the `{report-uri}` value in the default - this is cached as-is,\nwith the local report URL injected into it at runtime.\n',
    'author': 'YunoJuno',
    'author_email': 'code@yunojuno.com',
    'maintainer': 'YunoJuno',
    'maintainer_email': 'code@yunojuno.com',
    'url': 'https://github.com/yunojuno/django-csp-plus',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import logging
from collections import defaultdict
from typing import Any, TypeAlias

from django.db import models
from django.db.models import F
from django.db.utils import IntegrityError
from django.utils.timezone import now as tz_now
from pydantic import BaseModel, Field, root_validator, validator

from .settings import PolicyType
from .utils import strip_query

logger = logging.getLogger(__name__)

PydanticValues: TypeAlias = dict[str, str | None]


class ReportData(BaseModel):
    # browser support for CSP reports turns out to be patchy at best -
    # all fields are optional on the way in, but we need at least the
    # violated_directive and the blocked_uri to be able to make sense of
    # the report.

    # mandatory fields - without these we cannot process the report the
    # min_length ensures we don't have an empty string
    blocked_uri: str = Field(alias="blocked-uri", min_length=1)
    # we must have one of these - validate_directives enforces this
    effective_directive: str | None = Field(alias="effective-directive")
    violated_directive: str | None = Field(alias="violated-directive")
    # optional
    disposition: str | None = Field("", alias="disposition")
    document_uri: str | None = Field("", alias="document-uri")
    original_policy: str | None = Field(alias="original-policy")
    referrer: str | None = Field(alias="referrer")
    script_sample: str | None = Field(alias="script-sample")
    status_code: str | None = Field(0, alias="status-code")

    @validator("document_uri", "blocked_uri")
    def strip_uri(cls, uri: str) -> str:
        """
        Strip querystring and truncate to fit model length.

        We don't care about querystring params (CSP doesn't), and we can't
        store URLs > 200 chars long, so we truncate here.

        """
        return strip_query(uri)[:200] if uri else ""

    @root_validator
    def validate_directives(cls, values: PydanticValues) -> PydanticValues:
        """Ensure that we have either effective_directive or violated_directive."""
        if values.get("effective_directive", ""):
            return values
        # if effective_directive is empty, but violated_directive is not,
        # then udpate the former with the latter.
        if violated_directive := values.get("violated_directive", ""):
            logger.debug(
                "'effective_directive' missing - using 'violated_directive' attr."
            )
            values["effective_directive"] = violated_directive
            return values
        raise ValueError(
            "Either 'effective_directive' or 'violated_directive' must be present."
        )

    class Config:
        allow_population_by_field_name = True


class DispositionChoices(models.TextChoices):
    ENFORCE = ("enforce", "Enforce")
    REPORT = ("report", "Report only")


class DirectiveChoices(models.TextChoices):
    BASE_URI = ("base-uri", "base-uri")
    CHILD_SRC = ("child-src", "child-src")
    CONNECT_SRC = ("connect-src", "connect-src")
    DEFAULT_SRC = ("default-src", "default-src")
    FONT_SRC = ("font-src", "font-src")
    FORM_ACTION = ("form-action", "form-action")
    FRAME_ANCESTORS = ("frame-ancestors", "frame-ancestors")
    FRAME_SRC = ("frame-src", "frame-src")
    IMG_SRC = ("img-src", "img-src")
    MANIFEST_SRC = ("manifest-src", "manifest-src")
    MEDIA_SRC = ("media-src", "media-src")
    OBJECT_SRC = ("object-src", "object-src")
    PREFETCH_SRC = ("prefetch-src", "prefetch-src [DEPRECATED]")
    REPORT_TO = ("report-to", "report-to")
    REPORT_URI = ("report-uri", "report-uri")
    SCRIPT_SRC = ("script-src", "script-src")
    SCRIPT_SRC_ATTR = ("script-src-attr", "script-src-attr [CAUTION]")
    SCRIPT_SRC_ELEM = ("script-src-elem", "script-src-elem [CAUTION]")
    STYLE_SRC = ("style-src", "style-src")
    STYLE_SRC_ATTR = ("style-src-attr", "style-src-attr [CAUTION]")
    STYLE_SRC_ELEM = ("style-src-elem", "style-src-elem [CAUTION]")
    WORKER_SRC = ("worker-src", "worker-src")


class CspRuleQuerySet(models.QuerySet):
    def enabled(self) -> CspRuleQuerySet:
        return self.filter(enabled=True)

    def directive_values(self) -> models.ValuesQuerySet:
        return self.values_list("directive", "value")


class CspRuleManager(models.Manager):
    pass


class CspRule(models.Model):
    REQUIRE_SINGLE_QUOTE = [
        "nonce",
        "none",
        "report-sample",
        "self",
        "strict-dynamic",
        "unsafe-eval",
        "unsafe-hashes",
        "unsafe-inline",
        "wasm-unsafe-eval",
    ]
    REQUIRE_TRAILING_COLON = [
        "http",
        "https",
        "wss",
        "blob",
        "data",
        "mediastream",
        "filesystem",
    ]
    # require the "unsafe-" prefix
    REQUIRE_UNSAFE_PREFIX = ["inline", "eval"]

    directive = models.CharField(max_length=50, choices=DirectiveChoices.choices)
    value = models.CharField(max_length=255)
    enabled = models.BooleanField(default=False)
    created_at = models.DateTimeField(default=tz_now)
    modified_at = models.DateTimeField(default=tz_now)

    objects = CspRuleManager.from_queryset(CspRuleQuerySet)()

    class Meta:
        verbose_name = "CSP Rule"
        unique_together = ("value", "directive")
        ordering = ["directive", "value"]

    def __str__(self) -> str:
        return f"{self.directive} {self.value}"

    def save(self, *args: Any, **kwargs: Any) -> None:
        self.modified_at = tz_now()
        if "update_fields" in kwargs:
            kwargs["update_fields"].append("modified_at")
        super().save(*args, **kwargs)

    @classmethod
    def clean_value(cls, value: str) -> str:
        value = value.lower()
        if value in cls.REQUIRE_SINGLE_QUOTE:
            return f"'{value}'"
        if value in cls.REQUIRE_TRAILING_COLON:
            return f"{value}:"
        if value in cls.REQUIRE_UNSAFE_PREFIX:
            return f"'unsafe-{value}'"
        if source := strip_query(value):
            return source
        # not sure what we have here - let it go through
        return value


class CspReportQuerySet(models.QuerySet):
    pass


class CspReportManager(models.Manager):
    def save_report(self, data: ReportData) -> CspReport | None:
        report, _ = CspReport.objects.get_or_create(
            effective_directive=data.effective_directive,
            blocked_uri=data.blocked_uri,
        )
        # we udpate with the latest page that has caused the violation
        report.document_uri = data.document_uri
        report.disposition = data.disposition
        report.request_count = F("request_count") + 1
        report.last_updated_at = tz_now()
        report.save()
        return report


class CspReport(models.Model):
    # {
    #     'csp-report': {
    #         'document-uri': 'http://127.0.0.1:8000/test/',
    #         'referrer': '',
    #         'violated-directive': 'img-src',
    #         'effective-directive': 'img-src',
    #         'original-policy': "default-src https:; img-src 'self';",
    #         'disposition': 'enforce',
    #         'blocked-uri': 'https://example.com',
    #         'line-number': 8,
    #         'source-file': 'http://127.0.0.1:8000/test/',
    #         'status-code': 200,
    #         'script-sample': ''
    #     }
    # }
    document_uri = models.URLField()
    effective_directive = models.TextField()
    disposition = models.CharField(max_length=12)
    blocked_uri = models.URLField()
    request_count = models.IntegerField(default=0)
    created_at = models.DateTimeField(default=tz_now)
    last_updated_at = models.DateTimeField(default=tz_now)

    objects = CspReportManager.from_queryset(CspReportQuerySet)()

    class Meta:
        verbose_name = "CSP Violation"
        unique_together = ("effective_directive", "blocked_uri")
        ordering = ["effective_directive", "blocked_uri"]

    def __str__(self) -> str:
        return (
            f"CSP violation: {self.effective_directive} - "
            f"{self.blocked_uri} [{self.request_count}]"
        )


def convert_report(report: CspReport, enable: bool = True) -> CspRule | None:
    """Convert report to a rule and deletion the violation."""
    logger.debug("Converting violation report to new rule.")
    try:
        value = CspRule.clean_value(report.blocked_uri)
        rule = CspRule.objects.create(
            directive=report.effective_directive,
            value=value,
            enabled=enable,
        )
    except IntegrityError:
        # duplicate rule
        logger.debug("Duplicate rule found, deleting violation report.")
        report.delete()
        return None
    else:
        # once we have saved the rule, we delete the report
        logger.debug("Rule created, deleting violation report.")
        report.delete()
        return rule


class CspReportBlacklistQueryset(models.QuerySet):
    def as_dict(self) -> PolicyType:
        values = defaultdict(list)
        for directive, blocked_uri in self.values_list("directive", "blocked_uri"):
            values[directive].append(blocked_uri)
        return values


class CspReportBlacklist(models.Model):
    """
    Combination of directive + domain to ignore.

    CSP reports are flakey, to say the least, and some directive/domain
    combos get stuck. This model creates a blacklist that is used to
    stop logging reports once you are sure you have the rules set up.

    """

    directive = models.CharField(max_length=50, choices=DirectiveChoices.choices)
    blocked_uri = models.URLField()

    objects = CspReportBlacklistQueryset().as_manager()

    class Meta:
        verbose_name_plural = verbose_name = "CSP Blacklist"
        unique_together = ("directive", "blocked_uri")
        ordering = ["directive", "blocked_uri"]

    def __str__(self) -> str:
        return f"{self.directive} {self.blocked_uri}"



import base64
import logging
import os
import random
from functools import partial
from typing import Callable

from django.core.exceptions import MiddlewareNotUsed
from django.http import HttpRequest, HttpResponse
from django.utils.functional import SimpleLazyObject

from .policy import get_csp
from .settings import (
    CSP_ENABLED,
    CSP_REPORT_SAMPLING,
    CSP_RESPONSE_HEADER,
    REPORT_TO_HEADER,
    REPORTING_ENDPOINTS_HEADER,
    process_request,
    process_response,
)

logger = logging.getLogger(__name__)


def add_report_uri() -> bool:
    """Return True if we should add the report-uri directive."""
    return random.random() <= CSP_REPORT_SAMPLING  # noqa: S311


class CspNonceMiddleware:
    """Add the csp_nonce to all HttpResponses."""

    def __init__(self, get_response: Callable) -> None:
        if not CSP_ENABLED:
            raise MiddlewareNotUsed("Disabling CSPMiddleware")
        self.get_response = get_response

    def __call__(self, request: HttpRequest) -> HttpResponse | None:
        # direct lift from mozilla/django-csp (h/t)
        nonce = partial(self._make_nonce, request)
        request.csp_nonce = SimpleLazyObject(nonce)
        response = self.get_response(request)
        return response

    def _make_nonce(self, request: HttpRequest) -> str:
        if not getattr(request, "_csp_nonce", None):
            request._csp_nonce = base64.b64encode(os.urandom(16)).decode("ascii")
        return request._csp_nonce


class CspHeaderMiddleware:
    """Set the CSP header on the response."""

    def __init__(self, get_response: Callable) -> None:
        if not CSP_ENABLED:
            raise MiddlewareNotUsed("Disabling CSPMiddleware")
        self.get_response = get_response

    def __call__(self, request: HttpRequest) -> HttpResponse | None:
        response: HttpResponse = self.get_response(request)
        if not (process_request(request) and process_response(response)):
            return response
        self.add_csp_header(request, response)
        self.add_reporting_headers(response)
        return response

    def add_csp_header(self, request: HttpRequest, response: HttpResponse) -> None:
        response.headers[CSP_RESPONSE_HEADER] = get_csp(request, add_report_uri())

    def add_reporting_headers(self, response: HttpResponse) -> None:
        if REPORT_TO_HEADER:
            response.headers["Report-To"] = REPORT_TO_HEADER
        if REPORTING_ENDPOINTS_HEADER:
            response.headers["Reporting-Endpoints"] = REPORTING_ENDPOINTS_HEADER
